/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/

#include "datafox/communication/module_application.h"
#include "common/software_version.h"
#include <stdio.h>
#include <boost/algorithm/string.hpp>
#include <regex>

#include "spdlog/spdlog.h"
#include "spdlog/fmt/bin_to_hex.h"

// ##################################################################
/**
 * @brief Initialisieren der Membervariablen.
 *
 * @param id Nummer der Kanalklasse fuer z.B. Trace.
 */
Datafox::Communication::ModuleApplication::ModuleApplication( int32_t connectionNumber, uint8_t deviceAddress )
    : ModuleSystem( connectionNumber, deviceAddress ) 
{
    trace( "%s(%i):INFO(3): %s [ %i, %u ]\n", __FILE__, __LINE__, __func__, connectionNumber, deviceAddress );

    // Anzahl anstehender Datensaetze
    std::srand( (uint32_t)time( nullptr ) );

    // Callback fuer Nachricht auf Datensaetze
    m_callBackNotifyRecordAvailable = nullptr;
}

// ##################################################################
/**
 * @brief Destruktor.
 *
 */
Datafox::Communication::ModuleApplication::~ModuleApplication() 
{
    while( m_records.size() ) 
    {
        m_records.pop();
    }
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
void Datafox::Communication::ModuleApplication::errorModule( uint16_t module, uint16_t error, std::string &text )
{
    trace( "%s(%i):INFO(3): %s [ %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, module, error, text.c_str() );

    // Ist IV_MODULE_APPLICATION ?
    if ( module == IV_MODULE_APPLICATION )
    {

    }

    // Ermittlung des Fehlertextes an Basisklasse abgeben.
    ModuleSystem::errorModule( module, error, text );
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
void Datafox::Communication::ModuleApplication::errorFunction( uint16_t module, uint16_t func, uint16_t error, std::string &text )
{
    trace( "%s(%i):INFO(3): %s [ %u, %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, module, func, error, text.c_str() );

    if ( module == IV_MODULE_APPLICATION )
    {
        // Sonderfehler für Passwortschutz abhandeln.
        if (error == 28) 
        {
            // Die Modulfunktionen liefern den Fehlercode 28, wenn kein oder ein falsches Passwort verwendet wurde.
            // Es ist in diesem Fall zuerst das richtige Passwort für die Kommunikation erforderlich.
            text = boost::str( boost::format( "( %u ): Korrektes Passwort für die Kommunikation erforderlich.") % error );
            return;
        }
        else if (error == 29) 
        {
            // Die Modulfunktionen liefern den Fehlercode 29, wenn 3 mal ein falsches Passwort verwendet wurde.
            // Die Befehlsausführung ist danach für 60 Sekunden blockiert.
            text = boost::str( boost::format( "( %u ): Gerät ist aufgrund von fehlerhaften Passworteingaben für 60 Sekunden gesperrt.") % error );
            return;
        }

        switch (func)
        {
        case IV_FUNCTION_CLEAR_RECORDS:
        case IV_FUNCTION_GET_RECORD:
        case IV_FUNCTION_SET_RECORD:
            switch(error) {
            case 1:text = boost::str( boost::format( "( %u ): ") % error ); break;
            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!") % error ); break;
            }
            break;
        case IV_FUNCTION_SHOW_MESSAGE:
            switch(error) {
            case 1:text = boost::str( boost::format( "( %u ): Nachrichtentext zu lang fuer die Anzeige.") % error ); break;
            case 2:text = boost::str( boost::format( "( %u ): Es wird bereits eine Nachricht angezeigt.") % error ); break;
            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!") % error ); break;
            }
            break;
        case IV_FUNCTION_SET_VARIABLE:
        case IV_FUNCTION_GET_VARIABLE:
            switch(error) {
            case 1:text = boost::str( boost::format( "( %u ): Der Majorbestandteil des Variablennamens ist unbekannt.") % error ); break;
            case 2:text = boost::str( boost::format( "( %u ): Der Minorbestandteil des Variablennamens ist unbekannt.") % error ); break;
            case 3:
                text = boost::str( boost::format( "( %u ): Der verwendete Variablenwert ist zu lang.") % error );
                break;

            case 4:
                text = boost::str( boost::format( "( %u ): Der verwendete Variablentype würde nicht eingehalten.") % error );
                break;

            case 5:
                text = boost::str( boost::format( "( %u ): Das Schreib-/Leserecht würde nicht eingehalten.") % error );
                break;

            case 6:
                text = boost::str( boost::format( "( %u ): Das verwendete Format oder der Wertebereich würde nicht eingehalten.") % error );
                break;
            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!") % error ); break;
            }
            break;
        case IV_FUNCTION_SHOW_INFO:
            switch(error) {
            case 1:text = boost::str( boost::format( "( %u ): Nachrichtentext zu lang fuer die Anzeige.") % error ); break;
            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!") % error ); break;
            }
            break;
        case IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND:
        case IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND_OK:
            switch(error) {
            case 1: text = boost::str( boost::format( "( %u ): ZK nicht aktiv.") % error ); break;
            case 2: text = boost::str( boost::format( "( %u ): Parameter Command außerhalb seines gültigen Bereichs.") % error ); break;
            case 3: text = boost::str( boost::format( "( %u ): Parameter Master ID außerhalb seines gültigen Bereichs.") % error ); break;
            case 4: text = boost::str( boost::format( "( %u ): Parameter KnobId außerhalb seines gültigen Bereichs.") % error ); break;
            case 5: text = boost::str( boost::format( "( %u ): Speicherfehler.") % error ); break;
            case 6: text = boost::str( boost::format( "( %u ): Kommunikationsfehler mit dem Funkknoten.") % error ); break;
            case 7: text = boost::str( boost::format( "( %u ): Die Ausführung des Befehls erfolgte, doch der Schließlinder konnte nicht entfernt werden.") % error ); break;
            case 8: text = boost::str( boost::format( "( %u ): Die Ausführung des Befehls erfolgte, doch der Schließlinder konnte nicht eingelernt werden.") % error ); break;
            case 9: text = boost::str( boost::format( "( %u ): Die Ausführung des Befehls erfolgte, doch der Schließlinder konnte nicht eingelernt werden, weil er einen Produktschutz hat.") % error ); break;
            case 10: text = boost::str( boost::format( "( %u ): Das anzusprechende Modul ist nicht vorhanden." ) % error ); break; // Beim Cenloc, wenn auf dem Platz kein Modul bestückt wurde.
            case 11: text = boost::str( boost::format( "( %u ): Kein U&Z Funkmodul im Zutrittsbus vorhanden." ) % error ); break;
            default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Funktionsfehler!") % error ); break;
            }
            break;
        default: text = boost::str( boost::format( "( %u ): Nicht eingetragener Fehlerblock für Funktion!") % error ); break;
        }
        
        return;
    }

    // Ermittlung des Fehlertextes an Basisklasse abgeben.
    ModuleSystem::errorFunction( module, func, error, text );
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param deviceAddress Anzusprechende Geräteadresse.
 */
int32_t Datafox::Communication::ModuleApplication::remoteCall( uint16_t module, uint16_t func, uint8_t *data, uint16_t &count )
{
    if (module == IV_MODULE_APPLICATION)
    {
        uint8_t status;
        std::string param;
        switch (func)
        {
        case IV_FUNCTION_CLEAR_RECORDS: return clearRecords(deviceNumberSender());
        case IV_FUNCTION_SET_RECORD: return quitRecord(deviceNumberSender(), status, param);
        case IV_FUNCTION_GET_RECORD: return readRecord(deviceNumberSender(), data, count, status);
        case IV_FUNCTION_SHOW_MESSAGE: return writeMessageToDisplay(deviceNumberSender(), data[0], data[1], data[2], (char*)&data[3]);
        case IV_FUNCTION_SET_VARIABLE:
        {
            std::string name(reinterpret_cast<char *>(&data[0]));
            std::string val(reinterpret_cast<char *>(&data[name.size() + 1]));
            return writeSetupVariable(deviceNumberSender(), name, val);
        }
        case IV_FUNCTION_GET_VARIABLE:
        {
            std::string name(reinterpret_cast<char *>(&data[0]));
            std::string val;
            int32_t res = readSetupVariable(deviceNumberSender(), name, val);
            if (res == IV_ERROR_SUCCESS)
            {
                std::copy(val.cbegin(), val.cend(), &data[name.size() + 1]);
                count = static_cast<uint16_t>( name.size() + 1 + val.size() + 1 );
            }

            return res;
        }
        case IV_FUNCTION_SHOW_INFO: return writeMemoToDisplay(deviceNumberSender(), data[0], data[1], data[2], data[3], *(short*)&data[6], (char*)&data[8]);
        case IV_FUNCTION_ENTRANCE2_IDENTIFICATION: 
            {
            char *id, *pin;
            char status[64]; 
            int32_t statusLength;
            id = (char*)&data[strlen((char*)&data[0]) + 1];
            pin = &id[strlen(id) + 1];
            return accessControlIdentification(deviceNumberSender(), (char*)&data[0], id, pin, status, statusLength); 
            }
        case IV_FUNCTION_ENTRANCE2_IDENTIFICATION_OK: return replyAccessControlIdentification(deviceNumberSender());
        case IV_FUNCTION_NOTIFY: return notify(deviceNumberSender(), data[0], *(short*)&data[1], (char*)&data[3], (char*)&data[3 + strlen((char*)&data[3])]);
        case IV_FUNCTION_ACCESS_CONTROL_IDENTIFICATION_OK: return replyAccessControlIdentificationExt( deviceNumberSender() );
        case IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND_OK: return replyAccessControlKnobCommand( deviceNumberSender() );
        case IV_FUNCTION_CUSTOM_COMMAND_OK: return replyCustomCommand( deviceNumberSender() );
		// Funktion nicht definiert
        default: return 0x0200; 
        }

    }

    return ModuleSystem::remoteCall(module, func, data, count);
}

//#########################################################
//##################### APPLICATION #######################
//#########################################################

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param deviceAddress Anzusprechende Geräteadresse.
 */
int32_t Datafox::Communication::ModuleApplication::clearRecords( uint8_t deviceAddress )
{
    trace( "%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, deviceAddress );

    int32_t result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceAddress ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lockGuardCall guard( m_mutexCall );

        // Umgebung zuruecksetzen
        resetError();

        // Remoteadresse setzen.
        protocol().setDeviceNumberReceiver( deviceAddress );

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_CLEAR_RECORDS, m_packetData, count )) != IV_ERROR_SUCCESS ) 
        {
            // Fehler ausgeben
            traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_CLEAR_RECORDS, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if ( count ) 
        {
            trace( "%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten. ]\n", __FILE__, __LINE__, __func__, count );
        }
    }
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        uint8_t tmp[4];
        count = 0;

        while (m_records.size())
        {
            m_records.pop();
        }

        // Antwort des Befehls senden.
        if ( (result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_CLEAR_RECORDS, tmp, count)) != IV_ERROR_SUCCESS )
        {

            trace( "%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*
* @param deviceAddress Anzusprechende Geräteadresse.
*/
int32_t Datafox::Communication::ModuleApplication::restoreRecords( uint8_t deviceAddress )
{
    trace("%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, deviceAddress);

    int32_t result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceAddress ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lockGuardCall guard( m_mutexCall );

        // Umgebung zuruecksetzen
        resetError();

        // Remoteadresse setzen.
        setDeviceNumberReceiver(deviceAddress);

        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_RESTORE_RECORDS, m_packetData, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_RESTORE_RECORDS, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten. ]\n", __FILE__, __LINE__, __func__, count);
        }
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        uint8_t tmp[4];
        count = 0;

//		while (m_records.size())
//		{
//			m_records.pop();
//		}

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_RESTORE_RECORDS, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param deviceAddress Anzusprechende Geräteadresse.
 */
int32_t Datafox::Communication::ModuleApplication::quitRecord( uint8_t deviceAddress, uint8_t &status, std::string &param )
{
    trace("%s(%i):INFO(3): %s [ %u, %u ]\n", __FILE__, __LINE__, __func__, deviceAddress, status);

    int32_t result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceAddress ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lockGuardCall guard( m_mutexCall );

        // Umgebung zuruecksetzen
        resetError();

        // Remoteadresse setzen.
        setDeviceNumberReceiver(deviceAddress);

        // Wenn ein Parameterstring angegeben wurde, dann diesen weiterreichen. 
        // In der Antwort kann ein Antwortparameter enthalten sein, dieser wird dann als out zurückgegeben.
        if ( param.size() && param.size() < IV_PACKET_LENGTH_COMMAND_PARAMS )
        {
            std::copy( param.cbegin(), param.cend(), m_packetData );
            count += static_cast<uint16_t>( param.size() );
        }

        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_SET_RECORD, m_packetData, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_SET_RECORD, result );
            return result;
        }

        // OK
        if ( count > static_cast<uint16_t>( sizeof( uint8_t ) ) )
        {
            // Es wird von einem zusätzlichen Antwortparameter vom Typ String ausgegangen.
            if ( param.size() )
            {
                param.resize( count - sizeof( uint8_t ) );
                std::copy( &m_packetData[sizeof( uint8_t )], &m_packetData[count], std::back_inserter( param ) );
            }
            else
            {
                // Aktuell sollte ein Antwortparameter nur bei Datenpaketen mit Stringparamter in der Anfrage kommen.
                trace( "%s(%i):WARNING(1): %s [ Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count );
            }
        }
        
        switch ( m_packetData[0] ) 
        {
        case 0:
        case 2:
            status = 0;
            break;

        case MB_ANSW_ACK_ON_RECORD:        // Quittung erhalten (fuer Online-Datensatz) 
        case MB_ANSW_TIMEOUT_ACK_RECORD:   // Quittung kam nach Timeout (fuer Online-Datensatz) 
            status = 1;
            break;

        case MB_ANSW_OLD_ACK_ON_RECORD:    // wiederholt gesendete Quittung 
        case MB_ANSW_OLD_TIMEOUT_ACK_RECORD: // wiederholte Quittung nach Timeout 
            status = 2;
            break;

        case MB_ANSW_ACK_OFF_RECORD:       // Quittung erhalten (fuer Offline-Datensatz) 
            status = 3;
            break;

        case MB_ANSW_OLD_ACK_OFF_RECORD:   // wiederholt gesendete Quittung 
            status = 4;
            break;
        }
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        uint8_t data[1];

        if (m_records.size())
        {
            CRecord &record = m_records.front();

            switch (record.m_data[0])
            {
            case MB_ANSW_NO_RECORD:             // kein Datensatz vorhanden
                data[0] = 0;
                break;

            case MB_ANSW_ONLINE_RECORD:         // Online Datensatz
            case MB_ANSW_OLD_ONLINE_RECORD:     // wiederholt gesendeter Online Datensatz
                data[0] = MB_ANSW_ACK_ON_RECORD;
                break;

            case MB_ANSW_OFFLINE_RECORD:        // Offline Datensatz
            case MB_ANSW_OLD_OFFLINE_RECORD:    // wiederholt gesendeter Offline Datensatz
                data[0] = MB_ANSW_ACK_OFF_RECORD;
                break;
            }

            count = sizeof(uint8_t);

            // Datensatz loeschen
            m_records.pop();
        }
        else
        {
            data[0] = 0;
            count = sizeof(uint8_t);
        }

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_SET_RECORD, data, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param deviceAddress Anzusprechende Geräteadresse.
 */
int32_t Datafox::Communication::ModuleApplication::readRecord( uint8_t deviceAddress, uint8_t *buffer, uint16_t &count, uint8_t &status )
{
    trace("%s(%i):INFO(3): %s [ %u, %p, %u, %u ]\n", __FILE__, __LINE__, __func__, deviceAddress, buffer, count, status);
    int32_t result = IV_ERROR_SUCCESS;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceAddress ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lockGuardCall guard( m_mutexCall );

        // Mögliche Puffergröße für späteres kopieren der Daten halten.
        uint16_t l_count = count;

        // Fehler beim Abruf der Datensätze behandeln.
        enum class FailStatus
        {
            None,
            NoAnswerData,
            RecordStatusUnknown,
            StatusNoRecordButData,
            ReceivedDataToFew,
            ReceivedDataToMuch,
            RecordDescriptionOutOfRange,
            ReservedBufferToSmall
        } failStatus;
        bool swapToOffline;

label_skip:
        // Umgebung zuruecksetzen
        resetError();

        // Remoteadresse setzen.
        setDeviceNumberReceiver(deviceAddress);

        // Fehler beim Abruf der Datensätze behandeln.
        result = IV_ERROR_SUCCESS;
        status = 0;
        failStatus = FailStatus::None;
        swapToOffline = false;

        // Initalisieren der Empfangslänge.
        count = 0;

        if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_GET_RECORD, m_packetData, count )) != IV_ERROR_SUCCESS ) 
        {
            // Fehler ausgeben
            traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_GET_RECORD, result );
            return result;
        }

        // Bei der Verarbeitung muss sichergestellt werden, dass Datensätze nicht das weitere Auslesen behindern.
        // Es wird versucht, Datensätze die nicht verarbeitet werden können, zu sichern oder ggf. weiterzureichen.

        if ( count == 0 ) 
        {
            // Fehlerfall: Die Antwort ist leer.
            failStatus = FailStatus::NoAnswerData;
        }
        else
        {
            // Datensatzkennung nicht über _buffer_ durchreichen, wird über _status_ gemacht.
            --count;

            switch ( m_packetData[0] )
            {
                case MB_ANSW_NO_RECORD:             // kein Datensatz vorhanden
                    status = 0;
                    break;

                case MB_ANSW_ONLINE_RECORD:         // Online Datensatz 
                    status = 1;
                    break;

                case MB_ANSW_OLD_ONLINE_RECORD:     // wiederholt gesendeter Online Datensatz 
                    status = 2;
                    break;

                case MB_ANSW_OFFLINE_RECORD:        // Offline Datensatz 
                    status = 3;
                    break;

                case MB_ANSW_OLD_OFFLINE_RECORD:    // wiederholt gesendeter Offline Datensatz
                    status = 4;
                    break;

                default:
                {
                    // Fehlerfall: Die gelieferte Datensatzkennung ist unbekannt.
                    failStatus = FailStatus::RecordStatusUnknown;
                    break;
                }
            }

            if ( failStatus == FailStatus::None && status == 0 && count > 1 )
            {
                // Fehlerfall: Der Datensatzstatus ist "Kein Datensatz" es werden jedoch Daten geliefert.
                failStatus = FailStatus::StatusNoRecordButData;
            }

            if ( failStatus == FailStatus::RecordStatusUnknown || failStatus == FailStatus::StatusNoRecordButData )
            {
                // Wenn es sich um gültige Daten handeln könnte, diese zur Weiterverarbeitung freigeben.
                if ( count > 2 && count <= IV_RECORD_MAX && (m_packetData[1] < MAX_RECORD_DESCRIPTION) )
                {
                    // Es könnte sich um gültige Daten handeln. Die Verarebeitung wird der Anwendung überlassen.
                    // Die Daten werden als ein Offline-Datensatz weiterverarbeitet. Aber trotzdem geloggt.
                    status = 3;
                    swapToOffline = true;
                }
            }

            // Wurde Datensatz gemeldet
            if ( failStatus == FailStatus::None && status )
            {
                if ( count < 3 )
                {
                    // Datensatzbeschreibungsnummer + Leerfeld

                    // Fehlerfall: Gelieferter Datensatz ist kleiner als aktuell verarbeitbar / erwartet.
                    failStatus = FailStatus::ReceivedDataToFew;
                }
                else if ( count > IV_RECORD_MAX )
                {
                    // Fehlerfall: Gelieferter Datensatz ist größer als aktuell verarbeitbar / erwartet.
                    failStatus = FailStatus::ReceivedDataToMuch;
                }
                else if ( m_packetData[1] >= MAX_RECORD_DESCRIPTION )
                {
                    // Fehlerfall: Die Datensatzbeschreibungsnummer liegt außerhalb ihres aktuell gültigen Wertebereichs.
                    failStatus = FailStatus::RecordDescriptionOutOfRange;
                }
                else if ( count > l_count )
                {
                    // Fehlerfall: Gelieferter Datensatz passt nicht in den übergebenen Datenpuffer.
                    failStatus = FailStatus::ReservedBufferToSmall;
                }
            }
        }

        enum class LogStatus
        {
            None,
            TraceWarning,
            Warning,
            WarningWithData,
            Error,
            ErrorWithData
        }
        logStatus = LogStatus::None;
        std::string logMessage;

        switch ( failStatus )
        {
            case FailStatus::None:
                break;

            case FailStatus::NoAnswerData:
                // Dieser Fehler ist ein Protokollfehler der zum Abbruch der Verarbeitung führt.
                logStatus = LogStatus::Error;
                logMessage = "No response data to the record request.";

                // Verursacht ggf. ein "klemmen" des Datensatzabrufs und muss manuell behoben werden.
                result = -IV_ERROR_PROCESSING_DLL;
                break;

            case FailStatus::RecordStatusUnknown:
                // Fehlerfall: Die gelieferte Datensatzkennung ist unbekannt.
                logMessage = "The delivered record status is unknown.";
                if ( swapToOffline )
                {
                    // Scheinbar mitgesendete Daten werden als Offline-Datensatz weitergereicht.
                    logStatus = LogStatus::WarningWithData;
                    logMessage += " Set status to offline.";
                }
                else
                {
                    logStatus = LogStatus::ErrorWithData;
                }
                break;

            case FailStatus::StatusNoRecordButData:              
                // Fehlerfall: Der Datensatzstatus ist "Kein Datensatz" es werden jedoch Daten geliefert.
                logMessage = "The delivered record status is no data but data were delivered.";
                if ( swapToOffline )
                {
                    // Scheinbar mitgesendete Daten werden als Offline-Datensatz weitergereicht.
                    logStatus = LogStatus::WarningWithData;
                    logMessage += " Set status to offline.";
                }
                else
                {
                    logStatus = LogStatus::ErrorWithData;
                }
                break;

            case FailStatus::ReceivedDataToFew:
                // Fehlerfall: Gelieferter Datensatz ist kleiner als aktuell verarbeitbar / erwartet.
                logStatus = LogStatus::ErrorWithData;
                logMessage = "Supplied data set is smaller than currently processable / expected.";
                break;

            case FailStatus::ReceivedDataToMuch:
                // Fehlerfall: Gelieferter Datensatz ist größer als aktuell verarbeitbar / erwartet.
                // Dieser Fehler ist ein Protokollfehler der zum Abbruch der Verarbeitung führt.
                logStatus = LogStatus::ErrorWithData;
                logMessage = "The delivered data volume exceeds the current maximum processable size.";

                // Verursacht ggf. ein "klemmen" des Datensatzabrufs und muss manuell behoben werden.
                result = -IV_ERROR_PROCESSING_DLL;
                break;

            case FailStatus::RecordDescriptionOutOfRange:
                // Fehlerfall: Die Datensatzbeschreibungsnummer liegt außerhalb ihres aktuell gültigen Wertebereichs.
                logStatus = LogStatus::ErrorWithData;
                logMessage = "The delivered record structure number is larger than currently possible.";
                break;

            case FailStatus::ReservedBufferToSmall:
                // Fehlerfall: Gelieferter Datensatz passt nicht in den übergebenen Datenpuffer.
                // Zusätzlichen Trace-Eintrag machen.
                logStatus = LogStatus::TraceWarning;
                logMessage = "The delivered data volume exceeds the current maximum buffer size.";

                // Der gelieferte Datesatz kann nicht komplett in den zur Verfügung gestellten Datenpuffer kopiert werden.
                // Der zur Verfügung gestellte Datenpuffer muss vergrößert werden.
                result = -IV_ERROR_FUNCTION_PARAMETER;
                break;
        }

        if ( logStatus != LogStatus::None )
        {
            auto logger = spdlog::get( "recordProblemLogger" );
            if ( logger )
            {
                switch ( logStatus )
                {
                    case LogStatus::None:
                    case LogStatus::TraceWarning:
                        break;

                    case LogStatus::Warning:
                        logger->warn( logMessage );
                        break;

                    case LogStatus::WarningWithData:
                        logger->warn( logMessage + " {}",
                                        spdlog::to_hex( std::begin( m_packetData ), std::begin( m_packetData ) + count ) );
                        break;

                    case LogStatus::Error:
                        logger->error( logMessage );
                        break;

                    case LogStatus::ErrorWithData:
                        logger->error( logMessage + " {}",
                                        spdlog::to_hex( std::begin( m_packetData ), std::begin( m_packetData ) + count ) );
                        break;
                }
            }

            switch ( logStatus )
            {
                case LogStatus::TraceWarning:
                case LogStatus::Warning:
                case LogStatus::WarningWithData:
                    trace( "%s(%i):WARNING(1): %s [ %s ]\n", __FILE__, __LINE__, __func__, logMessage.c_str() );
                    break;

                 case LogStatus::Error:
                case LogStatus::ErrorWithData:
                    trace( "%s(%i):ERROR(1): %s [ %s ]\n", __FILE__, __LINE__, __func__, logMessage.c_str() );
                    break;
           }
        }

        if ( result == IV_ERROR_SUCCESS )
        {
            // Wenn der Datensatz nicht weitergereicht werden kann, muss er quittiert werden und der nächste angefragt.
            if ( failStatus != FailStatus::None && !swapToOffline )
            {
                uint8_t l_status = 0;
                std::string l_param;
                // Der Datensatz wird quittiert, geloggt und der nächste angefordert.
                if ( (result = quitRecord( deviceAddress, l_status, l_param )) != IV_ERROR_SUCCESS )
                {
                    // Fehler ausgeben
                    traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_SET_RECORD, result );
                    return result;
                }

                goto label_skip;
            }

            if ( status )
            {
                // Daten des Datensatzes in den Puffer kopieren.
                for ( uint16_t i = 0; i < count; i++ )
                {
                    buffer[i] = m_packetData[1 + i];
                }
            }
        }
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        uint8_t data[256];
        count = 0;

        if (m_records.size())
        {
            CRecord &record = m_records.front();

            // Daten kopieren
            data[0] = record.m_status;
            memcpy(&data[1], record.m_data, record.m_length);
            count = 1 + record.m_length;
        }
        else
        {
            data[count++] = MB_ANSW_NO_RECORD;
        }

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_GET_RECORD, data, (uint16_t &)count )) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param deviceAddress Anzusprechende Geräteadresse.
 */
int32_t Datafox::Communication::ModuleApplication::writeMessageToDisplay( uint8_t deviceAddress, uint8_t time, uint8_t action, uint8_t audio, const std::string &text )
{
    trace("%s(%i):INFO(3): %s [ %u, %u, %u, %u, %p ]\n", __FILE__, __LINE__, __func__, deviceAddress, time, action, audio, text.c_str());

    int32_t result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceAddress ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lockGuardCall guard( m_mutexCall );

        // Umgebung zuruecksetzen
        resetError();

        // Remoteadresse setzen.
        setDeviceNumberReceiver(deviceAddress);

        m_packetData[count] = time;
        count += sizeof(uint8_t);
        m_packetData[count] = action;
        count += sizeof(uint8_t);
        m_packetData[count] = audio;
        count += sizeof(uint8_t);

        // Text uebernehmen
        std::copy(text.cbegin(), text.cend(), &m_packetData[count]);
        count += static_cast<uint16_t>( text.size() );
        m_packetData[count] = 0; // Nullterminierung
        ++count;
        
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_SHOW_MESSAGE, m_packetData, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_SHOW_MESSAGE, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten. ]\n", __FILE__, __LINE__, __func__, count);
        }
    }
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        uint8_t tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_SHOW_MESSAGE, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        trace("%s(%i):INFO(1): %s [ %s ]\n", __FILE__, __LINE__, __func__, text.c_str());
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param deviceAddress Anzusprechende Geräteadresse.
 */
int32_t Datafox::Communication::ModuleApplication::writeSetupVariable( uint8_t deviceAddress, const std::string &name, const std::string &val )
{
    trace("%s(%i):INFO(3): %s [ %i, %s, %s ]\n", __FILE__, __LINE__, __func__, deviceAddress, name.c_str(), val.c_str());

    int32_t result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceAddress ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lockGuardCall guard( m_mutexCall );

        // Umgebung zuruecksetzen
        resetError();

        // Remoteadresse setzen.
        setDeviceNumberReceiver(deviceAddress);

        // Namen uebernehmen
        std::copy(name.cbegin(), name.cend(), &m_packetData[count]);
        count += static_cast<uint16_t>( name.size() );
        m_packetData[count] = 0; // Nullterminierung
        ++count;

        // Wert uebernehmen
        std::copy(val.cbegin(), val.cend(), &m_packetData[count]);
        count += static_cast<uint16_t>( val.size() );
        m_packetData[count] = 0;
        ++count;
        
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_SET_VARIABLE, m_packetData, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_SET_VARIABLE, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten. ]\n", __FILE__, __LINE__, __func__, count);
        }
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        uint8_t tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_SET_VARIABLE, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        trace("%s(%i):INFO(1): %s [ %s = %s ]\n", __FILE__, __LINE__, __func__, name.c_str(), val.c_str());
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param deviceAddress Anzusprechende Geräteadresse.
 */
int32_t Datafox::Communication::ModuleApplication::readSetupVariable( uint8_t deviceAddress, const std::string &name, std::string &val )
{
    trace("%s(%i):INFO(3): %s [ %i, %s, 0x%08X ]\n", __FILE__, __LINE__, __func__, deviceAddress, name.c_str(), val.c_str());

    int32_t result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceAddress ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lockGuardCall guard( m_mutexCall );

        // Umgebung zuruecksetzen
        resetError();

        // Remoteadresse setzen.
        setDeviceNumberReceiver(deviceAddress);

        // Namen uebernehmen
        std::copy(name.cbegin(), name.cend(), &m_packetData[count]);
        count += static_cast<uint16_t>( name.size() );
        m_packetData[count] = 0; // Nullterminierung
        ++count;
        
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_GET_VARIABLE, m_packetData, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_GET_VARIABLE, result );
            return result;
        }

        // Name+Wert als Antwortdaten erwartet
        if (count > (17 + 81)) 
        {
            trace("%s(%i):WARNING(1): %s [ Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count);
        }

        // Namen uebergehen
        count = 0;
        std::string lName(reinterpret_cast<char *>(&m_packetData[count]));
        count += static_cast<uint16_t>( lName.size() ) + 1;

        // OK
        val = reinterpret_cast<char *>(&m_packetData[count]);
    
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        uint8_t tmp[17 + 81];
        memset(tmp, 0, sizeof(tmp));
        count = static_cast<int32_t>( strlen( strncpy( (char*)tmp, name.c_str(), sizeof( tmp ) - 1 ) ) ) + 2;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_GET_VARIABLE, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        trace("%s(%i):INFO(1): %s [ %s ]\n", __FILE__, __LINE__, __func__, name.c_str());
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param deviceAddress Anzusprechende Geräteadresse.
 */
int32_t Datafox::Communication::ModuleApplication::writeMemoToDisplay( uint8_t deviceAddress, uint8_t maschine, uint8_t time, uint8_t font, uint8_t audio, uint16_t flags, const std::string &text )
{
    trace("%s(%i):INFO(3): %s [ %i, %i, %i, %i, %i, 0x%04X, %s ]\n", __FILE__, __LINE__, __func__, deviceAddress, maschine, time, font, audio, flags, text.c_str());

    int32_t result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
// Maschinennummer (1 Byte)
// Time ist die Anzeigedauer. (1 Byte) (0 im Hindergrund uebernehmen, -1 staendig anzeigen)
// Schriftgroessennummer (1 Byte)
// Akkustisches Signal (1 Byte)
// 16bit reserve (2 Byte)
// Flags z.b. 0x0001 (2 Byte) fuer Antiproportionale Schrift
// Infotext
    
    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceAddress ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lockGuardCall guard( m_mutexCall );

        // Umgebung zuruecksetzen
        resetError();

        // Remoteadresse setzen.
        setDeviceNumberReceiver(deviceAddress);

        m_packetData[count] = maschine;
        count += sizeof(uint8_t);
        m_packetData[count] = time;
        count += sizeof(uint8_t);
        m_packetData[count] = font;
        count += sizeof(uint8_t);
        m_packetData[count] = audio;
        count += sizeof(uint8_t);
        *reinterpret_cast<uint16_t *>( &m_packetData[count] ) = 0;
        count += sizeof( uint16_t );
        *reinterpret_cast<uint16_t *>( &m_packetData[count] ) = flags;
        count += sizeof( uint16_t );

        // Namen uebernehmen
        std::copy(text.cbegin(), text.cend(), &m_packetData[count]);
        count += static_cast<uint16_t>( text.size() );
        m_packetData[count] = 0; // Nullterminierung
        ++count;
        
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_SHOW_INFO, m_packetData, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_SHOW_INFO, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten. ]\n", __FILE__, __LINE__, __func__, count);
        }
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        uint8_t tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_SHOW_INFO, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        trace("%s(%i):INFO(1): %s [ %s ]\n", __FILE__, __LINE__, __func__, text.c_str());
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param deviceAddress Anzusprechende Geräteadresse.
 */
int32_t Datafox::Communication::ModuleApplication::accessControlIdentification( uint8_t deviceAddress, const char *tm, const char *id, const char *pin, char *status, int32_t &statusLength )
{
    trace("%s(%i):INFO(3): %s [ %i, %s, %s, %s ]\n", __FILE__, __LINE__, __func__, deviceAddress, tm, id, pin);

    int32_t result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceAddress ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lockGuardCall guard( m_mutexCall );

        std::string TM_( tm );
        std::string Id_( id );
        std::string Pin_( pin );
        std::string regTM("[0-3]\\d[01]");
        std::string regId("\\w{1,20}");
        std::string regPin("\\d{0,8}");

        // Format pruefen
        if ( !std::regex_match( TM_, std::regex( regTM ) ) 
            || !std::regex_match( Id_, std::regex( regId ) ) 
            || !std::regex_match( Pin_, std::regex( regPin ) ) )
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Umgebung zuruecksetzen
        resetError();

        // Remoteadresse setzen.
        setDeviceNumberReceiver(deviceAddress);

        // Namen uebernehmen
        count = static_cast<uint16_t>( strlen( strcpy( (char*)&m_packetData[count], tm ) ) ) + 1;
        count += static_cast<uint16_t>( strlen( strcpy( (char*)&m_packetData[count], id ) ) ) + 1;
        count += static_cast<uint16_t>( strlen( strcpy( (char*)&m_packetData[count], pin ) ) ) + 1;
        
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_ENTRANCE2_IDENTIFICATION, m_packetData, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_ENTRANCE2_IDENTIFICATION, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(1): %s [ Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count);
        }

        count = 0;

        if ((result = call(IV_CONTEXT_SLAVE, IV_MODULE_APPLICATION, IV_FUNCTION_ENTRANCE2_IDENTIFICATION_OK, m_packetData, count, IV_PENDING_TRIALS(connection().readTimeout()))) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_ENTRANCE2_IDENTIFICATION_OK, result );
            return result;
        }

        if (count == 0) 
        {
            trace("%s(%i):WARNING(1): %s [ Aktiver Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count);
        }

        // Wert von Status uebernehmen
        int32_t i;
        for (i=0; (i < count) && (i < statusLength); i++) 
        {
            status[i] = m_packetData[i];
        }
        // Uebergebene Laenge mitteilen
        statusLength = i - 1;
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        uint8_t tmp[64];
        count = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_ENTRANCE2_IDENTIFICATION, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        // Aktuellen Status als aktive Antwort senden
        count = static_cast<int32_t>( strlen( strcpy( (char*)tmp, "Keiner" ) ) ) + 1;

        // Aktive Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_ENTRANCE2_IDENTIFICATION_OK, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param deviceAddress Anzusprechende Geräteadresse.
 */
int32_t Datafox::Communication::ModuleApplication::replyAccessControlIdentification( uint8_t deviceAddress )
{
    trace("%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, deviceAddress);

    int32_t result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceAddress ) )
    {
#ifdef WITH_VIRTUAL_DEVICE
        trace("%s(%i):ERROR(1): %s \n", __FILE__, __LINE__, __func__);
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }
    else 
    {
        uint8_t tmp[4];
        count = 0;
        
        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_ENTRANCE2_IDENTIFICATION_OK, tmp, count)) != IV_ERROR_SUCCESS) 
        {	
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
    }

    return result;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*
* @param deviceAddress Anzusprechende Geräteadresse.
*/
int32_t Datafox::Communication::ModuleApplication::accessControlOnlineAction( uint8_t deviceAddress, const char *tm, uint32_t mask, uint8_t type, uint8_t duration )
{
    trace("%s(%i):INFO(3): %s [ %i, 0x08X, 0x%08X, %u, %u ]\n", __FILE__, __LINE__, __func__, deviceAddress, tm, mask, type, duration);

    int32_t result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceAddress ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lockGuardCall guard( m_mutexCall );

        std::string TM_( tm );
        std::string regTM("[0-3]\\d[01]");

        if ( !std::regex_match( TM_, std::regex( regTM ) ) ) 
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Umgebung zuruecksetzen
        resetError();

        // Remoteadresse setzen.
        setDeviceNumberReceiver(deviceAddress);

        // Nummer des Tuermodul uebernehmen
        count = static_cast<uint16_t>( strlen( strcpy( (char*)m_packetData, tm ) ) ) + 1;
        // Weiteren Werte uebernehmen
        *(uint32_t*)&m_packetData[count] = mask;
        count += sizeof(uint32_t);
        m_packetData[count] = type;
        count += sizeof(uint8_t);
        m_packetData[count] = duration;
        count += sizeof(uint8_t);

        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_ENTRANCE2_ONLINE_ACTION, m_packetData, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_ENTRANCE2_ONLINE_ACTION, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten. ]\n", __FILE__, __LINE__, __func__, count);
        }
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        uint8_t tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_ENTRANCE2_ONLINE_ACTION, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        trace("%s(%i):INFO(1): %s [ %s ]\n", __FILE__, __LINE__, __func__, tm);
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param deviceAddress Anzusprechende Geräteadresse.
 */
int32_t Datafox::Communication::ModuleApplication::notify( uint8_t deviceAddress, uint8_t type, int32_t priority, char *timestamp, char *params )
{
    trace("%s(%i):INFO(3): %s [ %i, %i, %i, %s ]\n", __FILE__, __LINE__, __func__, deviceAddress, type, priority, params);

    int32_t result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceAddress ) )
    {
#ifdef WITH_VIRTUAL_DEVICE
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lockGuardCall guard( m_mutexCall );

        // Umgebung zuruecksetzen
        resetError();

        // Remoteadresse setzen.
        setDeviceNumberReceiver(deviceAddress);

        // Benachrichtigungstyp uebernehmen
        m_packetData[count] = type;
        count += sizeof(uint8_t);
        // Prioritaet eintragen
        *(short*)&m_packetData[count] = priority;
        count += sizeof(short);
        // Zeitstempel
        if (timestamp == NULL) 
        {
            time_t tm_time;
            struct tm *ptm;
            tm_time = time(NULL);
            ptm = localtime(&tm_time);
            count += sprintf((char*)&m_packetData[count], "%02i%02i-%02i-%02i %02i:%02i:%02i:%03i", 
                (ptm->tm_year + 1900) / 100,
                (ptm->tm_year + 1900) % 100,
                ptm->tm_mon + 1,
                ptm->tm_mday,
                ptm->tm_hour,
                ptm->tm_min,
                ptm->tm_sec, 0) + 1;
        }
        else 
        {
            count += static_cast<int32_t>( strlen( strcpy( (char*)&m_packetData[count], timestamp ) ) ) + 1;
        }
        // Leerstring
        count += static_cast<int32_t>( strlen( strcpy( (char*)&m_packetData[count], params ) ) ) + 1;

        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_NOTIFY, m_packetData, count)) != IV_ERROR_SUCCESS)
        {
            // Fehler ausgeben
            traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_NOTIFY, result );
            return result;
        }

        // OK
        if (count > (int32_t)sizeof(int32_t)) 
        {
            trace("%s(%i):WARNING(1): %s [ Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count);
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }
    else 
    {
        uint8_t tmp[4];
        count = sizeof(tmp);
        tmp[0] = 1;

        switch (type)
        {
        case IV_VALUE_NOTIFY_RECORD:
            {
            CNotifyRecord nr;

            // Hat das Kanalobjekt den Aktiven Zustand?
            if (activeModeState() < ams_active)
            {
                // Funktion derzeit nicht ausführbar.
                return 0x0001;
            }
            //TRACE("ActiveState: %i\n", GetActiveState());

            if (m_callBackNotifyRecordAvailable)
            {
                int32_t pos = 0;

                // Aktuelle Kanalnummer eintragen
                nr.channelNumber = connectionNumber();

                // Aktuelle RemoteBusnummer eintragen
                nr.deviceAddress = deviceNumberReceiver();

                // Meldungstypen uebernehmen
                nr.type = type;
                pos += sizeof(uint8_t);

                // Meldungsprioritaet uebernehmen
                nr.priority = priority;
                pos += sizeof(uint16_t);

                // Zeitstempel der Meldungserzeugung uebernehmen
                pos += static_cast<int32_t>( strlen( strncpy( nr.timestamp, timestamp, sizeof( nr.timestamp ) - 1 ) ) );

                // Aktuellen Zeitstempel fuer Nachrichtenannahme halten
                nr.dt = time(NULL);

                // Meldung in Warteschlange stellen.
                if (m_callBackNotifyRecordAvailable(nr) != IV_ERROR_SUCCESS)
                {
                    // Die Annahme der Aktiven Datensatzmeldung schlug fehl (ggf. nicht aktiv).
                    return 0x0002;
                }

                trace("%s(%i):INFO(1): %s [ %i ]\n", __FILE__, __LINE__, __func__, connectionNumber());
            }
            else
            {
                // Die Annahme der Aktiven Datensatzmeldung ist nicht aktiv.
                return 0x0002;
            }
            break;
            }

        default:
            trace("%s(%i):INFO(1): %s [ Unknown type %i ]\n", __FILE__, __LINE__, __func__, type);
            break;
        }


        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_NOTIFY, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param deviceAddress Anzusprechende Geräteadresse.
 */
int32_t Datafox::Communication::ModuleApplication::notifyRecordAvailable( uint8_t deviceAddress, int32_t priority )
{
    trace("%s(%i):INFO(3): %s [ %i, %i ]\n", __FILE__, __LINE__, __func__, deviceAddress, priority);

    return notify(deviceAddress, IV_VALUE_NOTIFY_RECORD, priority, NULL, (char*)"");
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*
* @param deviceAddress Anzusprechende Geräteadresse.
* @param type 
0		Standardschrift(Sicherstellung der Abwärtskompatibilität)
1		16 Pixel (7Anzeigezeilen)
2		16 Pixel,feste breite(7Zeilen)
3		19 Pixel(6Anzeigezeilen)
4		19 Pixel,feste breite(6Zeilen)
5		21 Pixel(5Anzeigezeilen)
6		21 Pixel,feste breite(5Zeilen)
*/
int32_t Datafox::Communication::ModuleApplication::writeFontType( uint8_t deviceAddress, uint8_t type )
{
    trace("%s(%i):INFO(3): %s [ %i, %i ]\n", __FILE__, __LINE__, __func__, deviceAddress, type);

    int32_t result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceAddress ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lockGuardCall guard( m_mutexCall );

        // Umgebung zuruecksetzen
        resetError();

        // Remoteadresse setzen.
        setDeviceNumberReceiver(deviceAddress);

        // Typewert uebernehmen
        count = sizeof(uint8_t);
        // Wert uebernehmen
        *m_packetData = type;

        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_SET_TYPE_FONT, m_packetData, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_SET_TYPE_FONT, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten. ]\n", __FILE__, __LINE__, __func__, count);
        }
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        uint8_t tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_SET_TYPE_FONT, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
* @brief Ermittelt die Anzahl der vorliegenden Datensaetze oder verwendetes Speichervolumen.
*
* @param [in]     deviceAddress Anzusprechende Geräteadresse.
* @param [out]    number Parameter der die Anzahl Datensaetze erhaelt. Er wird auf -1 gesetzt, wenn der Wert nicht ermittelt werden konnte.
* @param [out]    memoryUsage Parameter der das Datenvolumen erhaelt. Er wird auf -1 gesetzt, wenn der Wert nicht ermittelt werden konnte.
* @param [out]    memorySize Parameter der das Gesamtvolumen fuer Datensaetze erhaelt. Er wird auf -1 gesetzt, wenn der Wert nicht ermittelt werden konnte.
*/
int32_t Datafox::Communication::ModuleApplication::readRecordsVolume( uint8_t deviceAddress, int32_t &number, int32_t &memoryUsage, int32_t &memorySize )
{
    trace("%s(%i):INFO(3): %s [ %i, %i, %i ]\n", __FILE__, __LINE__, __func__, deviceAddress, number, memoryUsage, memorySize);

    int32_t result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceAddress ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lockGuardCall guard( m_mutexCall );

        // Umgebung zuruecksetzen
        resetError();

        // Remoteadresse setzen.
        setDeviceNumberReceiver(deviceAddress);

        // Zuerst wird davon ausgegangen, dass beide Werte nicht ermittelt werden koennen.
        number = -1;
        memoryUsage = -1;
        memorySize = -1;

        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_RECORD_VOLUME, m_packetData, count)) == IV_ERROR_SUCCESS)
        {
            // Antwortdaten erwartet
            if (count != (int32_t)sizeof(uint32_t)*3)
            {
                trace("%s(%i):WARNING(1): %s [ Ermittlung Anzahl Datensaetze und Speicher mit %i Byte, ungleich erwartete %i Byte ]\n", __FILE__, __LINE__, __func__, count - 3 * sizeof(int32_t), 3 * sizeof(int32_t));
            }

            // Anzahl der Datensaetze uebernehmen.
            number = *(int32_t*)m_packetData;
            if (count > (int32_t)sizeof(int32_t))
                memoryUsage = *(int32_t*)&m_packetData[sizeof(int32_t)];
            if (count > (int32_t)sizeof(int32_t)*2)
                memorySize = *(int32_t*)&m_packetData[sizeof(int32_t)*2];

            return result;
        }

        // Nur der Modulfehler -Funktion unbekannt- wird akzeptiert.
        if (result < 0 || (result > 0 && ((result >> 8) & 0xFF) != 2)) {

            // Fehler ausgeben
            traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_RECORD_VOLUME, result );
            return result;

        }

        // Wird die Funktion nicht unterstuetzt, dann wird versucht ueber die Systemvariable SYSTEM.STATUS die Information
        // zu ermitteln.
        std::string status;
        if ((result = readSystemVariable(deviceNumberReceiver(), "SYSTEM.STATUS", status)) != IV_ERROR_SUCCESS) {

            return result;
        }

        // Im Status wird nach "RECORDS:" gesucht.
        // 1. Es liegt nicht vor.
        // 2. Es liegt mit der Angabe des benoetigten Datenvolumens vor.
        // 3. Es liegt mit der Angabe der Datensatzanzahl und des Datenvolumens vor.
        // 4. Bei aktueller Berechnung (Zwecks MMC) kann auch "RECORDS: ?" kommen, dann wird noch ermittelt.
        char *pos, *end, *res;
        int32_t val1 = -1; // Haelt den ersten ermittelten Wert.
        int32_t val2 = -1; // Haelt den zweiten ermittelten Wert
        pos = strstr(&status[0], "RECORDS:");
        if (pos == NULL) {

            // Die Informationen konnten nicht ermittelt werden.
            return result;
        }
        pos += 9;
        end = strstr(pos, "\r");
        if (end == NULL) end = &status[strlen(&status[0])];
        if ((end - pos) > 20 || (pos - &status[0]) < 20) {

            // Die Informationen konnten nicht ausgewertet werden.
            return result;
        }
        if (*pos == '?') {

            // Die Information wird intern noch ermittelt und steht zur Zeit noch nicht zur Verfuegung.
            return result;
        }
        // Die vorderen 20 Zeichen des status-Puffers fuer halten der Werte verwenden.
        res = &status[0];
        memset(res, 0, 20);
        while (pos != end) {

            if (isdigit(*pos)) {
                *res = *pos;
                res++;
            } else if (*pos == ' ') {
                // Wert 1 oder 2 fuellen.
                if (val1 == -1) {
                    val1 = atoi(&status[0]);
                    res = &status[0];
                    memset(res, 0, 20);
                } else if (val2 == -1) {
                    val2 = atoi(&status[0]);
                    break;
                }
            } else if (*pos == 'b') {
                // Es liegt nur der Erste Wert als Volumeangabe vor.
                break;
            }

            // Naechstes Zeichen.
            pos++;
        }

        if (val2 == -1) {
            // Es liegt die Angabe in Datenvolumen vor.
            memoryUsage = val1;
        } else {
            // Es liegen beide Angaben vor. Datensatzanzahl und volumen.
            number = val1;
            memoryUsage = val2;
        }
    }
    else
    {
#ifdef WITH_VIRTUAL_DEVICE
        uint8_t tmp[64];
        count = sizeof(int32_t)*3;
        *(int32_t*)tmp = '?';
        *(int32_t*)&tmp[sizeof(int32_t)] = 0;
        *(int32_t*)&tmp[sizeof(int32_t)*2] = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_RECORD_VOLUME, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param deviceAddress Anzusprechende Geräteadresse.
 */
int32_t Datafox::Communication::ModuleApplication::accessControlIdentificationExt( uint8_t deviceAddress, int32_t masterId, const char *tm, const char *id, const char *pin, char *status, int32_t &statusLength )
{
    trace("%s(%i):INFO(3): %s [ %i, %i, %s, %s, %s ]\n", __FILE__, __LINE__, __func__, deviceAddress, masterId, tm, id, pin);

    int32_t result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceAddress ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lockGuardCall guard( m_mutexCall );

        int32_t res;
        std::string buffer;
        std::string TM_( tm );
        std::string Id_( id );
        std::string Pin_( pin );
        std::string regTM("[0-3]\\d[01]");
        std::string regId("\\w{1,20}");
        std::string regPin("\\d{0,8}");

        // Format pruefen
        if ( !std::regex_match( TM_, std::regex( regTM ) ) 
            || !std::regex_match( Id_, std::regex( regId ) ) 
            || !std::regex_match( Pin_, std::regex( regPin ) ) )
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Angeggebene MasterId gegen die eingestellte prüfen.

        // Systemvariable ACCESS.MASTERID ermitteln.
        if ( (res = readSystemVariable( deviceAddress, "ACCESS.MASTERID", buffer )) != IV_ERROR_SUCCESS )
            return res;

        std::string currentMasterId( buffer );
        // Die anzusprechende MasterId muss größer oder gleich der aktuell eingestellten sein.
        if ( masterId < std::stoi( currentMasterId ) )
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Umgebung zuruecksetzen
        resetError();

        // Remoteadresse setzen.
        setDeviceNumberReceiver(deviceAddress);

        // Namen uebernehmen
        *reinterpret_cast<uint16_t*>( m_packetData ) = masterId;
        count = sizeof(uint16_t);
        count += static_cast<uint16_t>( strlen( strcpy( (char*)&m_packetData[count], tm ) ) ) + 1;
        count += static_cast<uint16_t>( strlen( strcpy( (char*)&m_packetData[count], id ) ) ) + 1;
        count += static_cast<uint16_t>( strlen( strcpy( (char*)&m_packetData[count], pin ) ) ) + 1;
        
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_IDENTIFICATION, m_packetData, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_IDENTIFICATION, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(1): %s [ Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count);
        }

        count = 0;

        if ((result = call(IV_CONTEXT_SLAVE, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_IDENTIFICATION_OK, m_packetData, count, IV_PENDING_TRIALS(connection().readTimeout()))) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_IDENTIFICATION_OK, result );
            return result;
        }

        if (count == 0) 
        {
            trace("%s(%i):WARNING(1): %s [ Aktiver Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count);
        }

        // Wert von Status uebernehmen
        int32_t i;
        for (i=0; (i < count) && (i < statusLength); i++) 
        {
            status[i] = m_packetData[i];
        }
        // Uebergebene Laenge mitteilen
        statusLength = i - 1;
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        uint8_t tmp[64];
        count = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_IDENTIFICATION, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        // Aktuellen Status als aktive Antwort senden
        count = static_cast<int32_t>( strlen( strcpy( (char*)tmp, "Keiner" ) ) ) + 1;

        // Aktive Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_IDENTIFICATION_OK, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param deviceAddress Anzusprechende Geräteadresse.
 */
int32_t Datafox::Communication::ModuleApplication::replyAccessControlIdentificationExt( uint8_t deviceAddress )
{
    trace("%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, deviceAddress);

    int32_t result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceAddress ) )
    {
#ifdef WITH_VIRTUAL_DEVICE
        trace("%s(%i):ERROR(1): %s \n", __FILE__, __LINE__, __func__);
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    } 
    else 
    {
        uint8_t tmp[4];
        count = 0;
        
        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_IDENTIFICATION_OK, tmp, count)) != IV_ERROR_SUCCESS) 
        {	
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
    }

    return result;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*
* @param deviceAddress Anzusprechende Geräteadresse.
*/
int32_t Datafox::Communication::ModuleApplication::accessControlOnlineActionExt( uint8_t deviceAddress, int32_t masterId, const char *tm, uint32_t mask, uint8_t type, uint8_t duration )
{
    trace("%s(%i):INFO(3): %s [ %i, 0x08X, 0x%08X, %u, %u ]\n", __FILE__, __LINE__, __func__, deviceAddress, tm, mask, type, duration);

    int32_t result = IV_ERROR_SUCCESS;
    uint16_t count = 0;

    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceAddress ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lockGuardCall guard( m_mutexCall );

        int32_t res;
        std::string buffer;
        std::string TM_( tm );
        std::string regTM("[0-3]\\d[01]");

        if ( !std::regex_match( TM_, std::regex( regTM ) ) ) 
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Angeggebene MasterId gegen die eingestellte prüfen.

        // Systemvariable ACCESS.MASTERID ermitteln.
        if ( (res = readSystemVariable( deviceAddress, "ACCESS.MASTERID", buffer )) != IV_ERROR_SUCCESS )
            return res;

        std::string currentMasterId( buffer );
        // Die anzusprechende MasterId muss größer oder gleich der aktuell eingestellten sein.
        if ( masterId < std::stoi( currentMasterId ) )
            return -IV_ERROR_FUNCTION_PARAMETER;

        // Umgebung zuruecksetzen
        resetError();

        // Remoteadresse setzen.
        setDeviceNumberReceiver(deviceAddress);

        // anzusprechende MasterId
        *reinterpret_cast<uint16_t*>( m_packetData ) = masterId;
        count = sizeof(uint16_t);
        // Nummer des Tuermodul uebernehmen
        count += static_cast<uint16_t>( strlen( strcpy( (char*)&m_packetData[count], tm ) ) ) + 1;
        // Weiteren Werte uebernehmen
        *(uint32_t*)&m_packetData[count] = mask;
        count += sizeof(uint32_t);
        m_packetData[count] = type;
        count += sizeof(uint8_t);
        m_packetData[count] = duration;
        count += sizeof(uint8_t);

        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_ONLINE_ACTION, m_packetData, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_ONLINE_ACTION, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(3): %s [ Antwort enthaelt %i Byte an unnuetze Daten. ]\n", __FILE__, __LINE__, __func__, count);
        }
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        uint8_t tmp[4];
        count = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_ONLINE_ACTION, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        trace("%s(%i):INFO(1): %s [ %s ]\n", __FILE__, __LINE__, __func__, tm);
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param deviceAddress Anzusprechende Geräteadresse.
 */
int32_t Datafox::Communication::ModuleApplication::accessControlKnobCommand( uint8_t deviceAddress, uint8_t knobCommand, uint16_t masterId, uint8_t knobId, const char *params, uint8_t *data, int32_t *sizeData )
{
    trace("%s(%i):INFO(3): %s [ %i, %i, %i, %i, %s ]\n", __FILE__, __LINE__, __func__, deviceAddress, knobCommand, masterId, knobId, params);

    int32_t result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceAddress ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lockGuardCall guard( m_mutexCall );

        // Umgebung zuruecksetzen
        resetError();

        // Remoteadresse setzen.
        setDeviceNumberReceiver(deviceAddress);

        // Namen uebernehmen
        *reinterpret_cast<uint8_t*>( &m_packetData[count] ) = knobCommand;
        count = sizeof(uint8_t);
        *reinterpret_cast<uint16_t*>( &m_packetData[count] ) = masterId;
        count += sizeof(uint16_t);
        *reinterpret_cast<uint8_t*>( &m_packetData[count] ) = knobId;
        count += sizeof(uint8_t);
        count += static_cast<uint16_t>( strlen( strcpy( (char*)&m_packetData[count], params ) ) ) + 1;
        
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND, m_packetData, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(1): %s [ Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count);
        }

        count = 0;

        if ((result = call(IV_CONTEXT_SLAVE, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND_OK, m_packetData, count, IV_PENDING_TRIALS(connection().readTimeout()))) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND_OK, result );
            return result;
        }

        // Die Daten werden als eine Listen von Hexwerten mit \r getrennt geliefert.
        // Keine Antwortdaten erwartet
        if ( knobCommand == IV_VALUE_ACCESS_CONTROL_KNOB_COMMAND_LIST && count != 64 ) 
        {
            trace( "%s(%i):WARNING(1): %s [ Es wurden 64 Byte Antwort erwartet und %i Byte erhalten. ]\n", __FILE__, __LINE__, __func__, count );
        }
        
        if ( knobCommand == IV_VALUE_ACCESS_CONTROL_KNOB_COMMAND_LIST )
        {
            // Die Daten müssen entsprechend aufbereitet werden.
            auto swapByte = []( uint8_t *d, uint8_t *s )
            {
                uint8_t tmp = *d;
                *d = *s;
                *s = tmp;
            };

            auto toLSB = [=]( uint8_t *p ) 
            { 
                swapByte( &p[0], &p[3] );
                swapByte( &p[1], &p[2] );
                return *reinterpret_cast<int32_t *>( p );
            };

            // Ggf. gelieferte Daten übernehmen.
            for ( int32_t i = 0; i < (count / 4) && ((i + 1) * 4 + i) < *sizeData; ++i )
                sprintf( reinterpret_cast<char *>( &data[i * 9] ), "%08X%c", toLSB( &m_packetData[i * 4] ), (i + 1) == (count / 4) ? '\0' : '\r' );

            *sizeData = count;
        }
        else
        {
            // Daten wie sie sind übernehmen.
            int32_t i = 0;
            for ( ; i < count && i < *sizeData; ++i )
                data[i] = m_packetData[i];

            *sizeData = i;
        }
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        uint8_t tmp[64];
        count = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        // Aktuellen Status als aktive Antwort senden
        count = static_cast<int32_t>( strlen( strcpy( (char*)tmp, "Keiner" ) ) ) + 1;

        // Aktive Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND_OK, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param deviceAddress Anzusprechende Geräteadresse.
 */
int32_t Datafox::Communication::ModuleApplication::replyAccessControlKnobCommand( uint8_t deviceAddress )
{
    trace("%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, deviceAddress);

    int32_t result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceAddress ) )
    {
#ifdef WITH_VIRTUAL_DEVICE
        trace("%s(%i):ERROR(1): %s \n", __FILE__, __LINE__, __func__);
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }
    else 
    {
        uint8_t tmp[4];
        count = 0;
        
        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_ACCESS_CONTROL_KNOB_COMMAND_OK, tmp, count)) != IV_ERROR_SUCCESS) 
        {	
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param deviceAddress Anzusprechende Geräteadresse.
 */
int32_t Datafox::Communication::ModuleApplication::customCommand( uint8_t deviceAddress, uint32_t identification, uint32_t customCommand, const char *dataIn, int32_t sizeIn, char *dataOut, int32_t *sizeOut )
{
    trace("%s(%i):INFO(3): %s [ %i, %i ]\n", __FILE__, __LINE__, __func__, deviceAddress, customCommand);

    int32_t result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceAddress ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lockGuardCall guard( m_mutexCall );

        // Umgebung zuruecksetzen
        resetError();

        // Remoteadresse setzen.
        setDeviceNumberReceiver(deviceAddress);
        
		*reinterpret_cast<uint32_t*>( &m_packetData[count] ) = identification;
		count = sizeof(uint32_t);
		*reinterpret_cast<uint32_t*>( &m_packetData[count] ) = customCommand;
		count += sizeof(uint32_t);

		*reinterpret_cast<uint32_t*>( &m_packetData[count] ) = sizeIn;
		count += sizeof(uint32_t);

		for ( int32_t i = 0; i < sizeIn; i ++ )
		{
			*reinterpret_cast<char*>( &m_packetData[count] ) =  dataIn[i];
			count += sizeof( char );
		}

        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_CUSTOM_COMMAND, m_packetData, count)) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_CUSTOM_COMMAND, result );
            return result;
        }

        // Keine Antwortdaten erwartet
        if (count) 
        {
            trace("%s(%i):WARNING(1): %s [ Antwortparameter (%i Byte) ]\n", __FILE__, __LINE__, __func__, count);
        }

        count = 0;

        if ((result = call(IV_CONTEXT_SLAVE, IV_MODULE_APPLICATION, IV_FUNCTION_CUSTOM_COMMAND_OK, m_packetData, count, IV_PENDING_TRIALS(connection().readTimeout()))) != IV_ERROR_SUCCESS) 
        {
            // Fehler ausgeben
            traceError( deviceAddress, IV_MODULE_APPLICATION, IV_FUNCTION_CUSTOM_COMMAND_OK,result );
            return result;
        }

        // Die Daten werden als eine Listen von Hexwerten mit \r getrennt geliefert.
        // Keine Antwortdaten erwartet

        // Daten wie sie sind übernehmen.
        int32_t i = 0;
        for ( ; i < count && i < *sizeOut; ++i )
            dataOut[i] = m_packetData[i];

        *sizeOut = i;
    } 
    else 
    {
#ifdef WITH_VIRTUAL_DEVICE
        uint8_t tmp[64];
        count = 0;

        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_CUSTOM_COMMAND, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }

        // Aktuellen Status als aktive Antwort senden
        count = static_cast<int32_t>( strlen( strcpy( (char*)tmp, "Keiner" ) ) ) + 1;

        // Aktive Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_MASTER, IV_MODULE_APPLICATION, IV_FUNCTION_CUSTOM_COMMAND_OK, tmp, count)) != IV_ERROR_SUCCESS)
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }

    return result;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 *
 * @param deviceAddress Anzusprechende Geräteadresse.
 */
int32_t Datafox::Communication::ModuleApplication::replyCustomCommand( uint8_t deviceAddress )
{
    trace("%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, deviceAddress);

    int32_t result = IV_ERROR_SUCCESS;
    uint16_t count = 0;
    
    // Funktion in Remoteklasse aufrufen?
    if ( !isOwnDeviceNumber( deviceAddress ) )
    {
#ifdef WITH_VIRTUAL_DEVICE
        trace("%s(%i):ERROR(1): %s \n", __FILE__, __LINE__, __func__);
#else
        result = 0x0200; // Funktion nicht verfügbar.
#endif
    }
    else 
    {
        uint8_t tmp[4];
        count = 0;
        
        // Antwort des Befehls senden.
        if ((result = call(IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_APPLICATION, IV_FUNCTION_CUSTOM_COMMAND_OK, tmp, count)) != IV_ERROR_SUCCESS) 
        {
            trace("%s(%i):ERROR(1): %s [ call (%i) ]\n", __FILE__, __LINE__, __func__, result);
            return result;
        }
    }

    return result;
}
