/**
* @file  df_log.cpp
* @brief Implementierung der Klasse CDFLog.
*/
//////////////////////////////////////////////////////////////////////

#include "log/df_log.h"

#include "file/dfcom_ini_helper.h"
#if defined(_MSC_VER) || defined(__MINGW32__)
#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif
#include <windows.h>
#include <direct.h>
#include <io.h>
#else
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#endif

#include <fstream>
#include <cstdio>
#include <cstdarg>
#include <cstring>
#include <array>
#include <sstream>
#include <thread>
#include <boost/algorithm/string.hpp>

const std::string CDFLog::m_logBeginFormatPure{ "%s\t%s%s" };
const std::string CDFLog::m_logBeginFormat{ "%s\t%s\t%i\t%i%s" };
const std::string CDFLog::m_logBeginFormatWithoutDeviceNumber{ "%s\t%s\t%i%s" };
const std::string CDFLog::m_logAddFormat{ "%s\t%s.%s\t%i\t%i\t%s" };
const std::string CDFLog::m_logAddFormatWithoutDeviceNumber{ "%s\t%s.%s\t%i\t%s" };

CDFLog::CDFLog( int connectionNumber, std::function<void( const std::string & )> changeTraceFilePath ) :
    callBackChangeTraceFilePath( changeTraceFilePath )
{
    // Zugewiesene Kanalnummer
    m_connectionNumber = connectionNumber;
    m_logCheck.m_connectionNumber = connectionNumber;

    // Bevorzugt auf die Dockingstation Type 2 prüfen.
    m_dockingPreferType = 2;

    // Das Oeffnen und Schliessen muss durch den selben Thread erfolgen.
    m_checkOpenAndCloseBySameThread = true;

    // Initialisierungsdatei lesen
    //logIniRead(); // führt auch logCheck durch.
}

CDFLog::~CDFLog()
{
}

#include <boost/algorithm/string.hpp>

void CDFLog::logIniRead()
{
    char section[20];
    std::string ini_file_path = DFComIniHelper::findIniFile();

    if ( !ini_file_path.size() )
    {
        // Fehler bei Ermittlung
        m_enabled = false;
        m_traceMode = 0; // Dann auch Trace deaktivieren.
        m_eraseFiles = false;
        m_writeList = 0;
        return;
    }

    std::sprintf( section, "Docking_%i", m_connectionNumber );
    // Bevorzugt Dockingstation V1/2 ansprechen.
    m_dockingPreferType = m_logIni.GetUInt( section, "Prefer", 1234, ini_file_path.c_str() );
    if ( m_dockingPreferType == 1234 ) // Ggf. globale Einstellung laden.
    {
        m_dockingPreferType = m_logIni.GetUInt( "Docking", "Prefer", 2, ini_file_path.c_str() );
    }

    if ( m_dockingPreferType < 1 )
    {
        m_dockingPreferType = 1;
    }
    else if ( m_dockingPreferType > 2 )
    {
        m_dockingPreferType = 2;
    }


    // Log aktiv?
    m_enabled = (m_logIni.GetUInt( "Log", "Enable", 1, ini_file_path.c_str() )) ? true : false;
    if ( !m_enabled )
    {
        // Abbruch durch Deaktivierung.
        m_traceMode = 0; // Dann auch Trace deaktivieren.
        m_eraseFiles = false;
        m_writeList = 0;
        return;
    }
    // Trace aktiv?
    m_traceMode = m_logIni.GetUInt( "Log", "Trace", 0, ini_file_path.c_str() );

    // Basispfad fuer Log
    std::memset( m_logCheck.m_basePath, 0, sizeof( m_logCheck.m_basePath ) );
    m_logIni.GetString( "Log", "Path", m_logCheck.m_basePath, ini_file_path.c_str() );
    if ( m_logCheck.m_basePath[0] == 0 ) 
    {
        std::string log_base_path( ini_file_path );
        auto pos = log_base_path.find_last_of( '/' );
        if ( pos != std::string::npos )
        {
            log_base_path.resize( pos );
        }
        // Verzeichnis der Ini als Basis
        std::strcpy( m_logCheck.m_basePath, log_base_path.c_str() );
    } 
    else 
    {
        std::string log_base_path( m_logCheck.m_basePath );
#if defined(_MSC_VER) || defined(__MINGW32__)
        // Windows akzeptiert auch '/' als Pfadtrenner, somit Gemeinsamkeit mit GNU/Linux herstellen.
        boost::replace_all( log_base_path, "\\", "/" );
#endif
        if ( '/' == log_base_path[log_base_path.size() - 1] )
        {
            log_base_path[log_base_path.size() - 1] = 0;
        }

        // Aufbereitetes Verzeichnis als Basis
        std::strcpy( m_logCheck.m_basePath, log_base_path.c_str() );
    }

    // Unterordner fuer Logdateien
    std::memset( m_logCheck.m_subFolder, 0, sizeof( m_logCheck.m_subFolder ) );
    m_logIni.GetString( "Log", "Folder", m_logCheck.m_subFolder, ini_file_path.c_str() );
    if ( m_logCheck.m_subFolder[0] == 0 )
    {
        std::strcpy( m_logCheck.m_subFolder, SUB_FOLDER );
    }

    std::string log_path( m_logCheck.m_basePath );
    log_path += '/';
    log_path += m_logCheck.m_subFolder;
    if ( '/' == log_path[log_path.size() - 1] )
    {
        log_path[log_path.size() - 1] = 0;
    }

    // pruefen und evnt. anlegen des Ordners
    struct stat lstat;
    if ( stat( log_path.c_str(), &lstat ) != 0 || (lstat.st_mode & S_IFMT) != S_IFDIR )
    {
#if defined(_MSC_VER) || defined(__MINGW32__)
        if ( mkdir( log_path.c_str() ) != 0 )
        {
#else
        if ( mkdir( log_path.c_str(), S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH) != 0 )
        {
#endif
            // Fehler bei Ermittlung
            m_enabled = false;
            m_traceMode = 0; // Dann auch Trace deaktivieren.
            m_eraseFiles = false;
            m_writeList = 0;
            return;
        }
    }

    // Schreibrecht prüfen.
    if ( access( log_path.c_str(), 2 ) == -1 )
    {
        // Fehler bei Ermittlung
        m_enabled = false;
        m_traceMode = 0; // Dann auch Trace deaktivieren.
        m_eraseFiles = false;
        m_writeList = 0;
        return;
    }

    // Alte Dateien loeschen
    m_eraseFiles = (m_logIni.GetUInt( "Log", "Erase", 1, ini_file_path.c_str() )) ? true : false;
    
    m_logCheck.m_eraseFileDays = m_logIni.GetUInt( "Log", "Keep", 10, ini_file_path.c_str() );
    if ( m_logCheck.m_eraseFileDays < 1 )
    {
        m_logCheck.m_eraseFileDays = 1;
    }
    if ( m_logCheck.m_eraseFileDays > 100 )
    {
        m_logCheck.m_eraseFileDays = 100;
    }

    // Ausfuehrungszeit angeben
    m_elapse = (m_logIni.GetUInt( "Log", "Elapse", 1, ini_file_path.c_str() )) ? true : false;
    // Listen wegschreiben
    m_writeList = m_logIni.GetUInt( "Log", "List", 1, ini_file_path.c_str() );

    // Sektion [Channels]
    // ----------------------
    // Oeffnen und Schliessen muss durch den selben Thread erfolgen.
    m_checkOpenAndCloseBySameThread = (m_logIni.GetUInt( "Channels", "CheckOpenCloseBySameThread", 1, ini_file_path.c_str() )) ? true : false;

    // Dateinamen für Log und Trace setzen, alte Logdateien ggf. loeschen.
    checkCurrentFilePaths();

    // read configured documentation settings (path, activation state, ...)
    std::array<char, 512> path{};
    m_logIni.GetString( "XML", "docPath", path.data(), ini_file_path.c_str() );
    m_docPath = path.data();
    m_createXMLHelpFile = (m_logIni.GetUInt( "XML", "createHelpFile", 0, ini_file_path.c_str() ) == 1);

    // read configured database settings (path, ...)
    path.fill( 0 );
    m_logIni.GetString( "DeviceUpdate", "dbPath", path.data(), ini_file_path.c_str() );
    m_dbPath = path.data();
    m_dbPathType = m_logIni.GetUInt( "DeviceUpdate", "dbPathType", 0, ini_file_path.c_str() );
}

void CDFLog::logCheck() 
{
    // Soll durchgefuehrt werden?
    if ( !m_enabled || !m_eraseFiles )
    {
        return;
    }

#if defined(_MSC_VER) || defined(__MINGW32__)
    /*
    // Suche nach Dateien
    sprintf(szFile, LOG_SEARCH_FILENAME, m_szBasePath, m_szSubFolder);
    intptr_t fp;
    struct finddata_t finddata;
    if ((fp = findfirst(szFile, &finddata)) != -1) {

    do {

    for (i=0; i < 5; i++) {
    // Stimmt mit einem der zu erhaltenden Dateien ueberein?
    if (strstr(szFileName[i], finddata.name) != 0) {
    break;
    }
    }
    if (i == 5) {
    // Sammeln der zu loeschenden Dateien,
    // da erst nach findclose geloescht werden kann!
    strcpy(szDeleteFileName[j++], finddata.name);
    } 

    } while ((j < 5) && (findnext(fp, &finddata) != -1));
    findclose(fp);

    if (j) {
    // Dateien loeschen
    for (i=0; i < j; i++) {

    sprintf(szFile, LOG_DELETE_FILENAME, m_szBasePath, m_szSubFolder, szDeleteFileName[i]);
    if (remove(szFile) == -1) {
    // Konnte nicht geloescht werden!
    }
    }
    }
    }
    */
    m_logCheck.check();

#else // 2011_02_03: VORSICHT: Unter Linux werden die Log-Dateien nicht gelöscht.
    /*
    int i, j;
    char szFile[260];
    char szFileName[5][260];
    char szDeleteFileName[5][260];

    // Namen der Logdateien fuer die letzten fuenf Tage erstellen.
    for (i=0; i < 5; i++) {
    // Dateiname der Logdatei
    time_t t = time(NULL);
    t -= 3600 * 24 * i;
    tm *ltm = localtime(&t);
    sprintf(szFileName[i], LOG_FILENAME, m_logCheck.m_szBasePath, m_logCheck.m_szSubFolder, 
    ltm->tm_year % 100, ltm->tm_mon + 1, ltm->tm_mday, m_channelID);

    //@LOG:
    printf("Halte Datei: %s\n", szFileName[i]);
    }
    // Alle anderen Logdateien loeschen 

    j = 0;

    // Suche nach Dateien
    sprintf(szFile, "%s%s", m_logCheck.m_szBasePath, m_logCheck.m_szSubFolder);

    DIR *dir;
    if ((dir = opendir(szFile)) != NULL) {

    struct dirent *ent;
    while((j < 5) && ((ent = readdir(dir)) != NULL)) {

    if (ent->d_type != DT_REG && ent->d_type != DT_UNKNOWN)
    {
    continue;
    }
    for (i=0; i < 5; i++) {
    // Stimmt mit einem der zu erhaltenden Dateien ueberein?
    if (strstr(szFileName[i], ent->d_name) != 0) {
    //@LOG:
    printf("Datei wird gehalten: %s\n", szFileName[i]);
    break;
    }
    }
    if (i == 5) {
    // Sammeln der zu loeschenden Dateien,
    // da erst nach closedir geloescht werden kann!
    strcpy(szDeleteFileName[j++], ent->d_name);
    } 
    }
    closedir(dir);

    if (j) {
    // Dateien loeschen
    for (i=0; i < j; i++) {

    sprintf(szFile, LOG_DELETE_FILENAME, m_logCheck.m_szBasePath, m_logCheck.m_szSubFolder, szDeleteFileName[i]);
    //@LOG:
    printf("Datei wird gelöscht: %s\n", szFile);
    if (remove(szFile) == -1) {
    // Konnte nicht geloescht werden!
    }
    }
    }
    }
    */
#endif

}

std::string CDFLog::currentLogFilePath()
{
    // Datum
    time_t t = std::time( nullptr );
    tm *ltm = localtime( &t );

    // Dateiname
    const char *basePath = m_logCheck.m_basePath;
    const char *subFolder = m_logCheck.m_subFolder;
    std::string filePath( std::snprintf( nullptr, 0, LOG_FILENAME, basePath, subFolder, ltm->tm_year % 100, ltm->tm_mon + 1, ltm->tm_mday, m_connectionNumber ) + 1, '\0' );
    std::snprintf( &filePath[0], filePath.size(), LOG_FILENAME, basePath, subFolder, ltm->tm_year % 100, ltm->tm_mon + 1, ltm->tm_mday, m_connectionNumber );

    return std::move( filePath );
}

std::string CDFLog::currentTraceFilePath()
{
    // Datum
    time_t t = std::time( nullptr );
    tm *ltm = localtime( &t );

    // Dateiname
    const char *basePath = m_logCheck.m_basePath;
    const char *subFolder = m_logCheck.m_subFolder;
    std::string filePath( std::snprintf( nullptr, 0, LOG_FILENAME, basePath, subFolder, ltm->tm_year % 100, ltm->tm_mon + 1, ltm->tm_mday, m_connectionNumber + 1000 ) + 1, '\0' );
    std::snprintf( &filePath[0], filePath.size(), LOG_FILENAME, basePath, subFolder, ltm->tm_year % 100, ltm->tm_mon + 1, ltm->tm_mday, m_connectionNumber + 1000 );

    return std::move( filePath );
}

void CDFLog::checkCurrentFilePaths()
{
    if ( currentLogFilePath() != m_filePath )
    {
        // Dateiname hat sich geändert.
        m_filePath = currentLogFilePath();

        if ( callBackChangeTraceFilePath != nullptr )
        {
            // Setzen des Namens fuer Tracefile
            callBackChangeTraceFilePath( currentTraceFilePath() );
        }

        logCheck(); // Eventuell die aelteste Logdatei loeschen.
    }
}

std::string CDFLog::entryPrefix( const time_t &t )
{
    uint32_t pid = 0;
    std::stringstream tid;
#if defined(_MSC_VER) || defined(__MINGW32__)
    pid = GetCurrentProcessId();
#else
    pid = getpid();
#endif
    tid << std::this_thread::get_id();

    // Uhrzeit des Logeintrages
    tm *ltm = std::localtime( &t );
    std::string entry( std::snprintf( nullptr, 0, "%02i:%02i:%02i\t%i\t%s", ltm->tm_hour, ltm->tm_min, ltm->tm_sec, pid, tid.str().c_str() ) + 1, '\0' );
    std::snprintf( &entry[0], entry.size(), "%02i:%02i:%02i\t%i\t%s", ltm->tm_hour, ltm->tm_min, ltm->tm_sec, pid, tid.str().c_str() );
    entry.resize( entry.size() - 1 ); // Reserviertes Nullbyte aus Größe entfernen.

    return std::move( entry );
}

// Beginnt einen Logeintrag für eine API-Funktion ohne vorgegebene Parameter wie Verbindungsnummer oder Gerätenummer auszugeben.
void CDFLog::logBeginPure( const char *functionName, const char *format, ... )
{
    if ( m_enabled == false )
    {
        return;
    }

    // Bereichssperre
    std::lock_guard<std::mutex> guard( m_mutexFile );

    // Prüfen der Dateinamen und ggf. anpassen.
    checkCurrentFilePaths();

    // Zusatzmeldungen verwerfen.
    if ( m_subEntries.size() )
    {
        m_subEntries.clear();
    }

    // Übernehmen der Grundinformation zum Logeintrag.
    m_functionName = functionName;
    m_deviceNumber = 0;

    // Vorsatz des Logeintrags; Uhrzeit, ProcessId, ThreadId
    std::string prefix( entryPrefix( std::time( nullptr ) ) );
    // Zeitpunkt des Begin für Berechnung der Ausführungsdauer
    m_clockBegin = std::clock();

    va_list ap;
    va_start( ap, format );

    std::string content( std::vsnprintf( nullptr, 0, format, ap ) + 1, '\0' );
    std::vsnprintf( &content[0], content.size(), format, ap );
    content.resize( content.size() - 1 ); // Reserviertes Nullbyte aus Größe entfernen.

    va_end( ap );

    std::string entry( std::snprintf( nullptr, 0, m_logBeginFormatPure.c_str(), prefix.c_str(), m_functionName.c_str(), content.c_str() ) + 1, '\0' );
    std::snprintf( &entry[0], entry.size(), m_logBeginFormatPure.c_str(), prefix.c_str(), m_functionName.c_str(), content.c_str() );
    entry.resize( entry.size() - 1 ); // Reserviertes Nullbyte aus Größe entfernen.

    FILE *Logfp = std::fopen( m_filePath.c_str(), "a" );
    if ( Logfp == nullptr )
    {
        return;
    }

    std::fwrite( entry.data(), 1, entry.size(), Logfp );

    std::fclose( Logfp );
}

// Beginnt einen Logeintrag für eine API-Funktion, gibt die zugeordnete Verbindungsnummer als ersten Parameter mit aus.
void CDFLog::logBegin( const char *functionName, const char *format, ... )
{
    if ( m_enabled == false )
    {
        return;
    }
        
    // Bereichssperre
    std::lock_guard<std::mutex> guard( m_mutexFile );

    // Prüfen der Dateinamen und ggf. anpassen.
    checkCurrentFilePaths();

    // Zusatzmeldungen verwerfen.
    if ( m_subEntries.size() )
    {
        m_subEntries.clear();
    }

    // Übernehmen der Grundinformation zum Logeintrag.
    m_functionName = functionName;
    m_deviceNumber = 0;

    // Vorsatz des Logeintrags; Uhrzeit, ProcessId, ThreadId
    std::string prefix( entryPrefix( std::time( nullptr ) ) );
    // Zeitpunkt des Begin für Berechnung der Ausführungsdauer
    m_clockBegin = std::clock();

    va_list ap;
    va_start( ap, format );

    std::string content( std::vsnprintf( nullptr, 0, format, ap ) + 1, '\0' );
    std::vsnprintf( &content[0], content.size(), format, ap );
    content.resize( content.size() - 1 ); // Reserviertes Nullbyte aus Größe entfernen.

    va_end( ap );

    std::string entry( std::snprintf( nullptr, 0, m_logBeginFormatWithoutDeviceNumber.c_str(), prefix.c_str(), m_functionName.c_str(), m_connectionNumber, content.c_str() ) + 1, '\0' );
    std::snprintf( &entry[0], entry.size(), m_logBeginFormatWithoutDeviceNumber.c_str(), prefix.c_str(), m_functionName.c_str(), m_connectionNumber, content.c_str() );
    entry.resize( entry.size() - 1 ); // Reserviertes Nullbyte aus Größe entfernen.

    FILE *Logfp = std::fopen( m_filePath.c_str(), "a" );
    if ( Logfp == nullptr )
    {
        return;
    }

    std::fwrite( entry.data(), 1, entry.size(), Logfp );

    std::fclose( Logfp );
}

// Beginnt einen Logeintrag für eine API-Funktion, gibt die zugeordnete Verbindungsnummer und Gerätenummer mit aus.
void CDFLog::logBegin( const char *functionName, int32_t deviceNumber, const char *format, ... )
{
    if ( m_enabled == false )
    {
        return;
    }

    // Bereichssperre
    std::lock_guard<std::mutex> guard( m_mutexFile );

    // Prüfen der Dateinamen und ggf. anpassen.
    checkCurrentFilePaths();

    // Zusatzmeldungen verwerfen.
    if ( m_subEntries.size() )
    {
        m_subEntries.clear();
    }

    // Übernehmen der Grundinformation zum Logeintrag.
    m_functionName = functionName;
    m_deviceNumber = deviceNumber;

    // Vorsatz des Logeintrags; Uhrzeit, ProcessId, ThreadId
    std::string prefix( entryPrefix( std::time( nullptr ) ) );
    // Zeitpunkt des Begin für Berechnung der Ausführungsdauer
    m_clockBegin = std::clock();

    va_list ap;
    va_start( ap, format );

    std::string content( std::vsnprintf( nullptr, 0, format, ap ) + 1, '\0' );
    std::vsnprintf( &content[0], content.size(), format, ap );
    content.resize( content.size() - 1 ); // Reserviertes Nullbyte aus Größe entfernen.

    va_end( ap );

    std::string entry;

    if ( m_deviceNumber > 0 )
    {
        entry.resize( std::snprintf( nullptr, 0, m_logBeginFormat.c_str(), prefix.c_str(), m_functionName.c_str(), m_connectionNumber, m_deviceNumber, content.c_str() ) + 1, '\0' );
        std::snprintf( &entry[0], entry.size(), m_logBeginFormat.c_str(), prefix.c_str(), m_functionName.c_str(), m_connectionNumber, m_deviceNumber, content.c_str() );
    }
    else
    {
        entry.resize( std::snprintf( nullptr, 0, m_logBeginFormatWithoutDeviceNumber.c_str(), prefix.c_str(), m_functionName.c_str(), m_connectionNumber, content.c_str() ) + 1, '\0' );
        std::snprintf( &entry[0], entry.size(), m_logBeginFormatWithoutDeviceNumber.c_str(), prefix.c_str(), m_functionName.c_str(), m_connectionNumber, content.c_str() );
    }

    entry.resize( entry.size() - 1 ); // Reserviertes Nullbyte aus Größe entfernen.

    FILE *Logfp = std::fopen( m_filePath.c_str(), "a" );
    if ( Logfp == nullptr )
    {
        return;
    }

    std::fwrite( entry.data(), 1, entry.size(), Logfp );

    std::fclose( Logfp );
}

// Führt einen Logeintrag einer API-Funktion fort.
void CDFLog::logContinue( const char *format, ... )
{
    if ( m_enabled == false )
    {
        return;
    }

    // Bereichssperre
    std::lock_guard<std::mutex> guard( m_mutexFile );

    FILE *Logfp = std::fopen( m_filePath.c_str(), "a");
    if ( Logfp == nullptr )
    {
        return;
    }

    va_list ap;
    va_start( ap, format );

    std::vfprintf( Logfp, format, ap );

    va_end( ap );

    std::fclose( Logfp );
}

// Beendet einen Logeintrag einer API-Funktion und schliesst mit CR ab.
void CDFLog::logEnd( const char *format, ... )
{
    if ( m_enabled == false )
    {
        return;
    }

    // Bereichssperre
    std::lock_guard<std::mutex> guard( m_mutexFile );

    FILE *Logfp = std::fopen( m_filePath.c_str(), "a");
    if ( Logfp == nullptr )
    {
        return;
    }

    va_list ap;
    va_start( ap, format );

    std::vfprintf( Logfp, format, ap );

    va_end( ap );

    // Ausfuehrungszeit angeben?
    if ( m_elapse )
    {
        // Ausfuehrungszeit berechnen
        double clockDiff;
        clockDiff = (double)(std::clock() - m_clockBegin) / CLOCKS_PER_SEC;
#if defined(_MSC_VER) || defined(__MINGW32__)
        std::fprintf( Logfp, "\t[%2.1f]\n", clockDiff );
#else
        std::fprintf( Logfp, "\t[%2.1f]\r\n", clockDiff );
#endif
    }
    else
    {
#if defined(_MSC_VER) || defined(__MINGW32__)
        std::fprintf( Logfp, "%s", "\n" );
#else
        std::fprintf( Logfp, "%s", "\r\n" );
#endif
    }

    // Zusatzmeldungen ausgeben.
    if ( m_subEntries.size() )
    {
        for ( auto &msg : m_subEntries )
        {
            std::fprintf( Logfp, msg.c_str() );
#if defined(_MSC_VER) || defined(__MINGW32__)
            std::fprintf( Logfp, "\n" );
#else
            std::fprintf( Logfp, "\r\n" );
#endif
        }
    }

    std::fclose( Logfp );
}

// Gibt zum aktuellen Logeintrag einer API-Funktion eine zusätzlich Informationzeile aus.
void CDFLog::logAddInfo( const char *format, ... )
{
    if ( m_enabled == false )
    {
        return;
    }

    va_list ap;
    va_start( ap, format );

    logAdd( "INFO", format, ap );

    va_end( ap );
}

// Gibt zum aktuellen Logeintrag einer API-Funktion eine zusätzlich Warnungszeile aus.
void CDFLog::logAddWarning( const char *format, ... )
{
    if ( m_enabled == false )
    {
        return;
    }

    va_list ap;
    va_start( ap, format );

    logAdd( "WARNING", format, ap );

    va_end( ap );
}

// Gibt zum aktuellen Logeintrag einer API-Funktion eine zusätzlich Fehlerzeile aus.
void CDFLog::logAddError( const char *format, ... )
{
    if ( m_enabled == false )
    {
        return;
    }

    va_list ap;
    va_start( ap, format );

    logAdd( "ERROR", format, ap );

    va_end( ap );
}

void CDFLog::logAdd( const char *addType, const char *format, va_list &va )
{
    if ( m_enabled == false )
    {
        return;
    }

    std::string prefix( entryPrefix( std::time( nullptr ) ) );

    va_list ap;
    va_copy( ap, va );

    std::string content( std::vsnprintf( nullptr, 0, format, ap ) + 1, '\0' );
    std::vsnprintf( &content[0], content.size(), format, ap );
    content.resize( content.size() - 1 ); // Reserviertes Nullbyte aus Größe entfernen.

    va_end( ap );
 
    std::string entry;

    if ( m_deviceNumber > 0 )
    {
        entry.resize( std::snprintf( nullptr, 0, m_logAddFormat.c_str(), prefix.c_str(), m_functionName.c_str(), addType, m_connectionNumber, m_deviceNumber, content.c_str() ) + 1, '\0' );
        std::snprintf( &entry[0], entry.size(), m_logAddFormat.c_str(), prefix.c_str(), m_functionName.c_str(), addType, m_connectionNumber, m_deviceNumber, content.c_str() );
    }
    else
    {
        entry.resize( std::snprintf( nullptr, 0, m_logAddFormatWithoutDeviceNumber.c_str(), prefix.c_str(), m_functionName.c_str(), addType, m_connectionNumber, content.c_str() ) + 1, '\0' );
        std::snprintf( &entry[0], entry.size(), m_logAddFormatWithoutDeviceNumber.c_str(), prefix.c_str(), m_functionName.c_str(), addType, m_connectionNumber, content.c_str() );
    }
    
    entry.resize( entry.size() - 1 ); // Reserviertes Nullbyte aus Größe entfernen.

    m_subEntries.push_back( std::move( entry ) );
}

#if defined(_MSC_VER) || defined(__MINGW32__)
void CDFLog::MakeFilePath( char *pathBuffer, char *fileName, char *fileExt )
{
    char drive[_MAX_DRIVE]{};
    char dir[_MAX_DIR]{};
    char fname[_MAX_FNAME]{};
    char ext[_MAX_EXT]{};

    _splitpath( pathBuffer, drive, dir, fname, ext );
    // Wenn keine Extension, dann gehoert eventueller Dateiname noch zum Pfad
    if ( !std::strlen( ext ) )
    {
        if ( std::strlen( fname ) )
        {
            std::strcat( pathBuffer, "\\" );
            // Neu auswerten
            _splitpath( pathBuffer, drive, dir, fname, ext );
        }
    }

    // Speicher loeschen
    std::memset( pathBuffer, 0, _MAX_PATH );

    _makepath( pathBuffer, drive, dir, (fileName) ? fileName : fname, (fileExt) ? fileExt : ext );
}
#endif


